/* vi: set ts=4 sw=4:*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
//#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include "abpmCpioFile.h"
#include "abpmRawFile.h"
#include "abpmWrongFile.h"
#include "abpmVector.h"
#include "abpmString.h"

struct cpio_hdr {		/* file member header */
	unsigned int magic;
	int dev;
	unsigned long ino;
	unsigned long mode;
	unsigned long uid;
	unsigned long gid;
	unsigned long nlink;
	int rdev;
	unsigned long mtime;
	unsigned long namesize;
	unsigned long filesize;
	unsigned long checksum;
	abpmString filename;
}; 

abpmCpioFile::abpmCpioFile(abpmString filename, bool useraw) {
	file_choosen = false;
	use_raw=useraw;
	raw_reader = new abpmVirtualFile(filename);
	find_type();
}

abpmCpioFile::abpmCpioFile(abpmFile *rawreader, bool useraw) {
	file_choosen = false;
	use_raw=useraw;
	raw_reader = rawreader;
	find_type();
}

void abpmCpioFile::find_type() {
	unsigned char buf[6];
	readbytes((char *)buf, 6);
	if(!strncmp((char *)buf, "070701", 6))
		type=NEW;
	else if(!strncmp((char *)buf, "070702", 6))
		type=NEWCRC;
	else if(!strncmp((char *)buf, "070707", 6))
		type=ASC;
	else if(buf[0]==0xc7 && buf[1]==0x71)
		type=BIN;
	else
		type=UNKNOWN;
	raw_reader->seek(0, SEEK_SET);
}

off_t abpmCpioFile::seek(off_t offset, int whence) {
	if(!file_choosen)
		return 0;
	off_t absolute_offset=0;
	if(whence==SEEK_CUR)
		absolute_offset = raw_reader->seek(0, SEEK_CUR) + offset;
	if(whence==SEEK_SET)
		absolute_offset = start + offset;
	if(offset==SEEK_END)
		absolute_offset = end + offset;
	if(absolute_offset > end)
		absolute_offset = end;
	if(absolute_offset < start)
		absolute_offset = start;
	off_t o_offset = raw_reader->seek(absolute_offset, SEEK_SET);
	return o_offset - start;
}

bool abpmCpioFile::eof() {
	if(!file_choosen)
		return true;
	off_t cur_pos = raw_reader->seek(0, SEEK_CUR);
	if(cur_pos >= end || cur_pos < start)
		return true;
	return false;
}

abpmVector<abpmString> abpmCpioFile::list() {
	abpmVector<abpmString> contents;
	struct cpio_hdr header;
	size_t offset;
	
	off_t o_offset = raw_reader->seek(0, SEEK_CUR);
	raw_reader->seek(0,SEEK_SET);

	while(readheader(&header)) {
		fprintf(stderr, "%s:%d: list: pos=%d, namesize=%d\n", __FILE__,__LINE__, raw_reader->seek(0, SEEK_CUR), header.namesize);
		contents.add(header.filename);
		offset = header.filesize + header.filesize % 2;
		offset +=raw_reader->seek(0, SEEK_CUR);
		if(type==NEW || type==NEWCRC)
			offset += offset % 4;
		fprintf(stderr, "%s:%d:offset = %d, = %d + %d\n", __FILE__,__LINE__,offset, header.filesize, header.filesize % 4);
		//offset = ((header.filesize % 2)==0 ? header.filesize : header.filesize+1);
		//offset = ((header.namesize % 2)==0 ? offset : offset+1);
		raw_reader->seek(offset, SEEK_SET);
	}
	raw_reader->seek(o_offset,SEEK_SET);
	return contents;
}

bool abpmCpioFile::is_file(abpmString path) {
	struct cpio_hdr header;
	size_t offset;
	int i;
	
	for(i=0; path[i]=='/'; i++);
	abpmString basename=path.tail(i);
	
	off_t o_offset = raw_reader->seek(0, SEEK_CUR);
	raw_reader->seek(0,SEEK_SET);

	while(readheader(&header)) {
		if(basename.eq(header.filename)) {
			raw_reader->seek(o_offset,SEEK_SET);
			return true;
		}
		offset = header.filesize + header.filesize % 4;
		//offset = ((header.namesize % 2)==0 ? header.filesize : header.filesize+1);
		//offset = ((offset % 2)==0 ? offset : offset+1);
		raw_reader->seek(offset, SEEK_CUR);
	}
	raw_reader->seek(o_offset,SEEK_SET);
	return false;
}

ssize_t abpmCpioFile::read(void *buf, size_t maxlen) {
	size_t real_maxlen = maxlen;
	off_t cur_pos = raw_reader->seek(0, SEEK_CUR);
	if(cur_pos >= end)
		return 0;
	if((off_t)(cur_pos + maxlen) > end)
		real_maxlen = end - cur_pos;
	return raw_reader->read(buf, real_maxlen);
}



/*returns a file descriptor that seeked to "file_to_search" file in ar file*/
//int ar_get_file(char *ar_filename, char *file_to_search, size_t *file_len);

bool abpmCpioFile::readheader(cpio_hdr *hd) {
	bool retval = false;
	switch(type) {
		case NEW:
		case NEWCRC:
			retval = readnewheader(hd);
			break;
		case ASC:
			retval = readascheader(hd);
			break;
		case BIN:
			retval = readbinheader(hd);
			break;
		default:
			break;
	}
	if(!retval)
		return false;
	char *filename=new char[hd->namesize];
	if(!readbytes(filename, hd->namesize)) {
		delete[] filename;
		return false;
	}
	fprintf(stderr, "%s:%d: namesize = %d, filename=%s\n", __FILE__,__LINE__,hd->namesize, filename);
	hd->filename = abpmString(filename, hd->namesize - 1);
	off_t offset = raw_reader->seek(0, SEEK_CUR) + hd->namesize % 2;
	if(type == NEW || type == NEWCRC)
		offset += offset % 4;
	raw_reader->seek(offset, SEEK_SET);
	return true;
}

bool abpmCpioFile::readbinheader(cpio_hdr *hd) {
	u_int16_t s1, s2, n[8];
	for(int i=0; i<8; i++)
		if(!readshort(&(n[i])))
			return false;
	hd->magic=n[0];
	hd->dev=n[1];
	hd->ino=n[2];
	hd->mode=n[3];
	hd->uid=n[4];
	hd->gid=n[5];
	hd->nlink=n[6];
	hd->rdev=n[7];
	if(!readshort(&s1))
		return false;
	if(!readshort(&s2))
		return false;
	hd->mtime = (((unsigned long)s1)<<16) + s2;
	if(!readshort(&s1))
		return false;
	hd->namesize = s1;
	if(!readshort(&s1))
		return false;
	if(!readshort(&s2))
		return false;
	hd->filesize = (((unsigned long)s1)<<16) + s2;
	return true;
}

bool abpmCpioFile::readnewheader(cpio_hdr *hd) {
	fprintf(stderr, "%s:%d: pos = %d\n", __FILE__,__LINE__, raw_reader->seek(0, SEEK_CUR));
	char buf[111];
	if(!readbytes(buf, 110))
		return false;
	buf[110]=0;
	unsigned long dev_min, dev_maj, rdev_min, rdev_maj;
	sscanf(buf, "%6o%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx",
			&(hd->magic),
			&(hd->ino),
			&(hd->mode),
			&(hd->uid),
			&(hd->gid),
			&(hd->nlink),
			&(hd->mtime),
			&(hd->filesize),
			&dev_maj,
			&dev_min,
			&rdev_maj,
			&rdev_min,
			&(hd->namesize),
			&(hd->checksum));
	fprintf(stderr, "%s: %d: readnewheader: \nbuf=\t%s\n"
			"read=\t%6o%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx%8lx\n",__FILE__,__LINE__,buf,
			(hd->magic),
			(hd->ino),
			(hd->mode),
			(hd->uid),
			(hd->gid),
			(hd->nlink),
			(hd->mtime),
			(hd->filesize),
			dev_maj,
			dev_min,
			rdev_maj,
			rdev_min,
			(hd->namesize),
			(hd->checksum));
	fprintf(stderr, "filesize = %d\n", hd->filesize);
	hd->dev=0x100*dev_maj+dev_min;
	hd->rdev=0x100*rdev_maj+rdev_min;
	return true;
}
bool abpmCpioFile::readascheader(cpio_hdr *hd) {
	char buf[12];
	unsigned short n[8];
	buf[6]='\0';
	for(int i=0; i<8; i++) {
		if(!readbytes(buf, 6))
			return false;
		sscanf(buf, "%ho", &(n[i]));
	}
	hd->magic=n[0];
	hd->dev=n[1];
	hd->ino=n[2];
	hd->mode=n[3];
	hd->uid=n[4];
	hd->gid=n[5];
	hd->nlink=n[6];
	hd->rdev=n[7];
	if(!readbytes(buf, 11))
		return false;
	buf[11]='\0';
	sscanf(buf, "%lo", &(hd->mtime));
	if(!readbytes(buf, 6))
		return false;
	buf[6]='\0';
	sscanf(buf, "%lo", &(hd->namesize));
	if(!readbytes(buf, 11))
		return false;
	buf[11]='\0';
	sscanf(buf, "%lo", &(hd->filesize));
	return true;
}

abpmFile *abpmCpioFile::choose_file(abpmString path) {
	struct cpio_hdr header;
	size_t offset;
	int i;
	
	for(i=0; path[i]=='/'; i++);
	abpmString basename=path.tail(i);

	file_choosen=false;
	
	while(readheader(&header)) {
		if(basename.eq(header.filename)) {
			file_choosen=true;
			return this;
		}
		offset = header.filesize + header.filesize % 4;
		//offset = ((header.namesize % 2)==0 ? header.filesize : header.filesize+1);
		//offset = ((offset % 2)==0 ? offset : offset+1);
		raw_reader->seek(offset, SEEK_CUR);
	}
	
	return new abpmWrongFile(this, abpmString(__FILE__)+": "+__LINE__+": ");
}
